@@ -84,6 +84,14 @@ typedef struct SMTP_ITERATOR {
 	vstring_strcpy((iter)->dest, STR((iter)->saved_dest)); \
     } while (0)
 
+#define SMTP_ITER_UPDATE_HOST(iter, _host, _addr, _rr) do { \
+	vstring_strcpy((iter)->host, (_host)); \
+	vstring_strcpy((iter)->addr, (_addr)); \
+	(iter)->rr = (_rr); \
+	if ((_rr)->port) \
+	    (iter)->port = htons((_rr)->port); /* SRV port override */ \
+    } while (0)
+
  /*
   * TLS Policy support.
   */
@@ -191,6 +199,12 @@ typedef struct SMTP_STATE {
      * assume per-server debug_peer support.
      */
     int     debug_peer_per_nexthop;
+
+    /*
+     * One-bit counters to avoid logging the same warning multiple times per
+     * delivery request.
+     */
+    int     logged_line_length_limit:1;
 } SMTP_STATE;
 
  /*
@@ -267,6 +281,7 @@ typedef struct SMTP_STATE {
 #define SMTP_MISC_FLAG_COMPLETE_SESSION	(1<<7)
 #define SMTP_MISC_FLAG_PREF_IPV6	(1<<8)
 #define SMTP_MISC_FLAG_PREF_IPV4	(1<<9)
+#define SMTP_MISC_FLAG_FALLBACK_SRV_TO_MX (1<<10)
 
 #define SMTP_MISC_FLAG_CONN_CACHE_MASK \
 	(SMTP_MISC_FLAG_CONN_LOAD | SMTP_MISC_FLAG_CONN_STORE)
@@ -306,6 +321,8 @@ extern MAPS *smtp_generic_maps;		/* make internal address valid */
 extern int smtp_ext_prop_mask;		/* address extension propagation */
 extern unsigned smtp_dns_res_opt;	/* DNS query flags */
 
+extern STRING_LIST *smtp_use_srv_lookup;/* services with SRV record lookup */
+
 #ifdef USE_TLS
 
 extern TLS_APPL_STATE *smtp_tls_ctx;	/* client-side TLS engine */
