@@ -11,6 +11,9 @@
 /*
 /*	void	smtpd_peer_reset(state)
 /*	SMTPD_STATE *state;
+/* AUXILIARY METHODS
+/*	void	smtpd_peer_from_default(state)
+/*	SMTPD_STATE *state;
 /* DESCRIPTION
 /*	The smtpd_peer_init() routine attempts to produce a printable
 /*	version of the peer name and address of the specified socket.
@@ -96,6 +99,10 @@
 /* .RE
 /* .PP
 /*	smtpd_peer_reset() releases memory allocated by smtpd_peer_init().
+/*
+/*	smtpd_peer_from_default() looks up connection information
+/*	when an up-stream proxy indicates that a connection is not
+/*	proxied.
 /* LICENSE
 /* .ad
 /* .fi
@@ -118,7 +125,6 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
-#include <stdio.h>			/* strerror() */
 #include <errno.h>
 #include <netdb.h>
 #include <string.h>
@@ -145,7 +151,7 @@
 
 #include "smtpd.h"
 
-static INET_PROTO_INFO *proto_info;
+static const INET_PROTO_INFO *proto_info;
 
  /*
   * XXX If we make local port information available via logging, then we must
@@ -258,7 +264,7 @@ static int smtpd_peer_sockaddr_to_hostaddr(SMTPD_STATE *state)
 	    /*
 	     * Following RFC 2821 section 4.1.3, an IPv6 address literal gets
 	     * a prefix of 'IPv6:'. We do this consistently for all IPv6
-	     * addresses that that appear in headers or envelopes. The fact
+	     * addresses that appear in headers or envelopes. The fact
 	     * that valid_mailhost_addr() enforces the form helps of course.
 	     * We use the form without IPV6: prefix when doing access
 	     * control, or when accessing the connection cache.
@@ -511,7 +517,7 @@ static void smtpd_peer_from_pass_attr(SMTPD_STATE *state)
 
 /* smtpd_peer_from_default - try to initialize peer information from socket */
 
-static void smtpd_peer_from_default(SMTPD_STATE *state)
+void    smtpd_peer_from_default(SMTPD_STATE *state)
 {
 
     /*
@@ -544,8 +550,7 @@ static void smtpd_peer_from_proxy(SMTPD_STATE *state)
 {
     typedef struct {
 	const char *name;
-	int     (*endpt_lookup) (SMTPD_STATE *,
-			            void (*default_lookup) (SMTPD_STATE *));
+	int     (*endpt_lookup) (SMTPD_STATE *);
     } SMTPD_ENDPT_LOOKUP_INFO;
     static const SMTPD_ENDPT_LOOKUP_INFO smtpd_endpt_lookup_info[] = {
 	HAPROXY_PROTO_NAME, smtpd_peer_from_haproxy,
@@ -564,8 +569,8 @@ static void smtpd_peer_from_proxy(SMTPD_STATE *state)
 	if (strcmp(var_smtpd_uproxy_proto, pp->name) == 0)
 	    break;
     }
-    if (pp->endpt_lookup(state, smtpd_peer_from_default) < 0) {
-	smtpd_peer_no_client(state);
+    if (pp->endpt_lookup(state) < 0) {
+	smtpd_peer_from_default(state);
 	state->flags |= SMTPD_FLAG_HANGUP;
     } else {
 	smtpd_peer_hostaddr_to_sockaddr(state);
