@@ -37,6 +37,7 @@
 /*	char	*var_tls_eecdh_auto;
 /*	char	*var_tls_eecdh_strong;
 /*	char	*var_tls_eecdh_ultra;
+/*	char	*var_tls_ffdhe_auto;
 /*	char	*var_tls_dane_digests;
 /*	int	var_tls_daemon_rand_bytes;
 /*	bool	var_tls_append_def_CA;
@@ -95,13 +96,15 @@
 /*	int	where;
 /*	int	ret;
 /*
-/*	long	tls_bio_dump_cb(bio, cmd, argp, argi, argl, ret)
+/*	long	tls_bio_dump_cb(bio, cmd, argp, len, argi, argl, ret, processed)
 /*	BIO	*bio;
 /*	int	cmd;
 /*	const char *argp;
+/*	size_t	len;
 /*	int	argi;
 /*	long	argl; /* unused */
-/*	long	ret;
+/*	int	ret;
+/*	size_t	*processed;
 /*
 /*	int	tls_log_mask(log_param, log_level)
 /*	const char *log_param;
@@ -283,6 +286,7 @@ int     var_tls_daemon_rand_bytes;
 char   *var_tls_eecdh_auto;
 char   *var_tls_eecdh_strong;
 char   *var_tls_eecdh_ultra;
+char   *var_tls_ffdhe_auto;
 char   *var_tls_dane_digests;
 bool    var_tls_append_def_CA;
 char   *var_tls_bug_tweaks;
@@ -647,9 +651,10 @@ void    tls_param_init(void)
 	VAR_TLS_LOW_CLIST, DEF_TLS_LOW_CLIST, &var_tls_low_clist, 1, 0,
 	VAR_TLS_EXPORT_CLIST, DEF_TLS_EXPORT_CLIST, &var_tls_export_clist, 1, 0,
 	VAR_TLS_NULL_CLIST, DEF_TLS_NULL_CLIST, &var_tls_null_clist, 1, 0,
-	VAR_TLS_EECDH_AUTO, DEF_TLS_EECDH_AUTO, &var_tls_eecdh_auto, 1, 0,
+	VAR_TLS_EECDH_AUTO, DEF_TLS_EECDH_AUTO, &var_tls_eecdh_auto, 0, 0,
 	VAR_TLS_EECDH_STRONG, DEF_TLS_EECDH_STRONG, &var_tls_eecdh_strong, 1, 0,
 	VAR_TLS_EECDH_ULTRA, DEF_TLS_EECDH_ULTRA, &var_tls_eecdh_ultra, 1, 0,
+	VAR_TLS_FFDHE_AUTO, DEF_TLS_FFDHE_AUTO, &var_tls_ffdhe_auto, 0, 0,
 	VAR_TLS_BUG_TWEAKS, DEF_TLS_BUG_TWEAKS, &var_tls_bug_tweaks, 0, 0,
 	VAR_TLS_SSL_OPTIONS, DEF_TLS_SSL_OPTIONS, &var_tls_ssl_options, 0, 0,
 	VAR_TLS_DANE_DIGESTS, DEF_TLS_DANE_DIGESTS, &var_tls_dane_digests, 1, 0,
@@ -861,18 +866,50 @@ const char *tls_set_ciphers(TLS_SESS_STATE *TLScontext, const char *grade,
     return (vstring_str(buf));
 }
 
+/* ec_curve_name - copy EC key curve group name */
+
+#ifndef OPENSSL_NO_EC
+static char *ec_curve_name(EVP_PKEY *pkey)
+{
+    char   *curve = 0;
+
+#if OPENSSL_VERSION_PREREQ(3,0)
+    size_t  namelen;
+
+    if (EVP_PKEY_get_group_name(pkey, 0, 0, &namelen)) {
+	curve = mymalloc(++namelen);
+	if (!EVP_PKEY_get_group_name(pkey, curve, namelen, 0)) {
+	    myfree(curve);
+	    curve = 0;
+	}
+    }
+#else
+    EC_KEY *eckey = EVP_PKEY_get0_EC_KEY(pkey);
+    int     nid = EC_GROUP_get_curve_name(EC_KEY_get0_group(eckey));
+    const char *tmp = EC_curve_nid2nist(nid);
+
+    if (!tmp)
+	tmp = OBJ_nid2sn(nid);
+    if (tmp)
+	curve = mystrdup(tmp);
+#endif
+    return (curve);
+}
+
+#endif
+
 /* tls_get_signature_params - TLS 1.3 signature details */
 
 void    tls_get_signature_params(TLS_SESS_STATE *TLScontext)
 {
     const char *kex_name = 0;
-    const char *kex_curve = 0;
     const char *locl_sig_name = 0;
-    const char *locl_sig_curve = 0;
     const char *locl_sig_dgst = 0;
     const char *peer_sig_name = 0;
-    const char *peer_sig_curve = 0;
     const char *peer_sig_dgst = 0;
+    char   *kex_curve = 0;
+    char   *locl_sig_curve = 0;
+    char   *peer_sig_curve = 0;
     int     nid;
     SSL    *ssl = TLScontext->con;
     int     srvr = SSL_is_server(ssl);
@@ -882,11 +919,6 @@ void    tls_get_signature_params(TLS_SESS_STATE *TLScontext)
     X509   *peer_cert;
     EVP_PKEY *peer_pkey = 0;
 
-#ifndef OPENSSL_NO_EC
-    EC_KEY *eckey;
-
-#endif
-
 #define SIG_PROP(c, s, p) (*((s) ? &c->srvr_sig_##p : &c->clnt_sig_##p))
 
     if (SSL_version(ssl) < TLS1_3_VERSION)
@@ -906,11 +938,7 @@ void    tls_get_signature_params(TLS_SESS_STATE *TLScontext)
 #ifndef OPENSSL_NO_EC
 	case EVP_PKEY_EC:
 	    kex_name = "ECDHE";
-	    eckey = EVP_PKEY_get0_EC_KEY(dh_pkey);
-	    nid = EC_GROUP_get_curve_name(EC_KEY_get0_group(eckey));
-	    kex_curve = EC_curve_nid2nist(nid);
-	    if (!kex_curve)
-		kex_curve = OBJ_nid2sn(nid);
+	    kex_curve = ec_curve_name(dh_pkey);
 	    break;
 #endif
 	}
@@ -951,11 +979,7 @@ void    tls_get_signature_params(TLS_SESS_STATE *TLScontext)
 #ifndef OPENSSL_NO_EC
 	    case EVP_PKEY_EC:
 		locl_sig_name = "ECDSA";
-		eckey = EVP_PKEY_get0_EC_KEY(local_pkey);
-		nid = EC_GROUP_get_curve_name(EC_KEY_get0_group(eckey));
-		locl_sig_curve = EC_curve_nid2nist(nid);
-		if (!locl_sig_curve)
-		    locl_sig_curve = OBJ_nid2sn(nid);
+		locl_sig_curve = ec_curve_name(local_pkey);
 		break;
 #endif
 	    }
@@ -970,7 +994,7 @@ void    tls_get_signature_params(TLS_SESS_STATE *TLScontext)
 	    locl_sig_dgst = OBJ_nid2sn(nid);
     }
     /* Signature algorithms for the peer end of the connection */
-    if ((peer_cert = SSL_get_peer_certificate(ssl)) != 0) {
+    if ((peer_cert = TLS_PEEK_PEER_CERT(ssl)) != 0) {
 	peer_pkey = X509_get0_pubkey(peer_cert);
 
 	/*
@@ -993,11 +1017,7 @@ void    tls_get_signature_params(TLS_SESS_STATE *TLScontext)
 #ifndef OPENSSL_NO_EC
 	    case EVP_PKEY_EC:
 		peer_sig_name = "ECDSA";
-		eckey = EVP_PKEY_get0_EC_KEY(peer_pkey);
-		nid = EC_GROUP_get_curve_name(EC_KEY_get0_group(eckey));
-		peer_sig_curve = EC_curve_nid2nist(nid);
-		if (!peer_sig_curve)
-		    peer_sig_curve = OBJ_nid2sn(nid);
+		peer_sig_curve = ec_curve_name(peer_pkey);
 		break;
 #endif
 	    }
@@ -1010,24 +1030,21 @@ void    tls_get_signature_params(TLS_SESS_STATE *TLScontext)
 	if (SSL_get_peer_signature_nid(ssl, &nid) && nid != NID_undef)
 	    peer_sig_dgst = OBJ_nid2sn(nid);
 
-	X509_free(peer_cert);
+	TLS_FREE_PEER_CERT(peer_cert);
     }
     if (kex_name) {
 	TLScontext->kex_name = mystrdup(kex_name);
-	if (kex_curve)
-	    TLScontext->kex_curve = mystrdup(kex_curve);
+	TLScontext->kex_curve = kex_curve;
     }
     if (locl_sig_name) {
 	SIG_PROP(TLScontext, srvr, name) = mystrdup(locl_sig_name);
-	if (locl_sig_curve)
-	    SIG_PROP(TLScontext, srvr, curve) = mystrdup(locl_sig_curve);
+	SIG_PROP(TLScontext, srvr, curve) = locl_sig_curve;
 	if (locl_sig_dgst)
 	    SIG_PROP(TLScontext, srvr, dgst) = mystrdup(locl_sig_dgst);
     }
     if (peer_sig_name) {
 	SIG_PROP(TLScontext, !srvr, name) = mystrdup(peer_sig_name);
-	if (peer_sig_curve)
-	    SIG_PROP(TLScontext, !srvr, curve) = mystrdup(peer_sig_curve);
+	SIG_PROP(TLScontext, !srvr, curve) = peer_sig_curve;
 	if (peer_sig_dgst)
 	    SIG_PROP(TLScontext, !srvr, dgst) = mystrdup(peer_sig_dgst);
     }
@@ -1355,6 +1372,12 @@ long    tls_bug_bits(void)
      * options just in case.
      */
     bits |= SSL_OP_SINGLE_ECDH_USE | SSL_OP_SINGLE_DH_USE;
+
+    /*
+     * Unconditionally disable a CPU resource attack. There's no good reason
+     * to enable TLS renegotiation in the middle of an SMTP connection.
+     */
+    bits |= SSL_OP_NO_RENEGOTIATION;
     return (bits);
 }
 
@@ -1369,7 +1392,14 @@ void    tls_print_errors(void)
     int     line;
     int     flags;
 
-    while ((err = ERR_get_error_line_data(&file, &line, &data, &flags)) != 0) {
+#if OPENSSL_VERSION_PREREQ(3,0)
+/* XXX: We're ignoring the function name, do we want to log it? */
+#define ERRGET(fi, l, d, fl) ERR_get_error_all(fi, l, 0, d, fl)
+#else
+#define ERRGET(fi, l, d, fl) ERR_get_error_line_data(fi, l, d, fl)
+#endif
+
+    while ((err = ERRGET(&file, &line, &data, &flags)) != 0) {
 	ERR_error_string_n(err, buffer, sizeof(buffer));
 	if (flags & ERR_TXT_STRING)
 	    msg_warn("TLS library problem: %s:%s:%d:%s:",
@@ -1432,7 +1462,7 @@ void    tls_info_callback(const SSL *s, int where, int ret)
   * 
   * Modified to save a lot of strcpy and strcat by Matti Aarnio.
   * 
-  * Rewritten by Wietse to elimate fixed-size stack buffer, array index
+  * Rewritten by Wietse to eliminate fixed-size stack buffer, array index
   * multiplication and division, sprintf() and strcpy(), and lots of strlen()
   * calls. We could make it a little faster by using a fixed-size stack-based
   * buffer.
@@ -1493,6 +1523,7 @@ static void tls_dump_buffer(const unsigned char *start, int len)
 
 /* taken from OpenSSL apps/s_cb.c */
 
+#if !OPENSSL_VERSION_PREREQ(3,0)
 long    tls_bio_dump_cb(BIO *bio, int cmd, const char *argp, int argi,
 			        long unused_argl, long ret)
 {
@@ -1510,28 +1541,50 @@ long    tls_bio_dump_cb(BIO *bio, int cmd, const char *argp, int argi,
     return (ret);
 }
 
+#else
+long    tls_bio_dump_cb(BIO *bio, int cmd, const char *argp, size_t len,
+	             int argi, long unused_argl, int ret, size_t *processed)
+{
+    size_t  bytes = (ret > 0 && processed != NULL) ? *processed : len;
+
+    if (cmd == (BIO_CB_READ | BIO_CB_RETURN)) {
+	if (ret > 0) {
+	    msg_info("read from %08lX [%08lX] (%ld bytes => %ld (0x%lX))",
+		     (unsigned long) bio, (unsigned long) argp, (long) len,
+		     (long) bytes, (long) bytes);
+	    tls_dump_buffer((unsigned char *) argp, (int) bytes);
+	} else {
+	    msg_info("read from %08lX [%08lX] (%ld bytes => %d)",
+		     (unsigned long) bio, (unsigned long) argp,
+		     (long) len, ret);
+	}
+    } else if (cmd == (BIO_CB_WRITE | BIO_CB_RETURN)) {
+	if (ret > 0) {
+	    msg_info("write to %08lX [%08lX] (%ld bytes => %ld (0x%lX))",
+		     (unsigned long) bio, (unsigned long) argp, (long) len,
+		     (long) bytes, (long) bytes);
+	    tls_dump_buffer((unsigned char *) argp, (int) bytes);
+	} else {
+	    msg_info("write to %08lX [%08lX] (%ld bytes => %d)",
+		     (unsigned long) bio, (unsigned long) argp,
+		     (long) len, ret);
+	}
+    }
+    return ret;
+}
+
+#endif
+
 const EVP_MD *tls_validate_digest(const char *dgst)
 {
     const EVP_MD *md_alg;
-    unsigned int md_len;
 
     /*
      * If the administrator specifies an unsupported digest algorithm, fail
      * now, rather than in the middle of a TLS handshake.
      */
-    if ((md_alg = EVP_get_digestbyname(dgst)) == 0) {
+    if ((md_alg = tls_digest_byname(dgst, NULL)) == 0)
 	msg_warn("Digest algorithm \"%s\" not found", dgst);
-	return (0);
-    }
-
-    /*
-     * Sanity check: Newer shared libraries may use larger digests.
-     */
-    if ((md_len = EVP_MD_size(md_alg)) > EVP_MAX_MD_SIZE) {
-	msg_warn("Digest algorithm \"%s\" output size %u too large",
-		 dgst, md_len);
-	return (0);
-    }
     return md_alg;
 }
 
